import copy

import numpy as np
import pytest

from openff.toolkit import Molecule, Topology
from openff.toolkit._tests.create_molecules import create_ethanol
from openff.toolkit._tests.create_molecules import (
    dipeptide_residues_perceived as create_dipeptide,
)
from openff.toolkit._tests.utils import get_data_file_path, requires_rdkit
from openff.toolkit.topology._mm_molecule import _SimpleAtom, _SimpleMolecule


@pytest.fixture()
def water():
    water = _SimpleMolecule()
    water.add_atom(atomic_number=8)
    water.add_atom(atomic_number=1)
    water.add_atom(atomic_number=1)
    water.add_bond(0, 1)
    water.add_bond(0, 2)

    return water


@pytest.fixture()
def methane():
    methane = _SimpleMolecule()
    methane.add_atom(atomic_number=6)
    methane.add_atom(atomic_number=1)
    methane.add_atom(atomic_number=1)
    methane.add_atom(atomic_number=1)
    methane.add_atom(atomic_number=1)
    methane.add_bond(0, 1)
    methane.add_bond(0, 2)
    methane.add_bond(0, 3)
    methane.add_bond(0, 4)

    return methane


@pytest.fixture()
def methanol():
    methanol = _SimpleMolecule()
    methanol.add_atom(atomic_number=8)
    methanol.add_atom(atomic_number=6)
    methanol.add_atom(atomic_number=1)
    methanol.add_atom(atomic_number=1)
    methanol.add_atom(atomic_number=1)
    methanol.add_bond(0, 1)
    methanol.add_bond(1, 2)
    methanol.add_bond(1, 3)
    methanol.add_bond(1, 4)

    return methanol


@pytest.fixture()
def molecule_with_zero_atom():
    molecule = _SimpleMolecule()
    molecule.add_atom(atomic_number=6)
    molecule.add_atom(atomic_number=0)
    molecule.add_bond(0, 1)

    return molecule


@pytest.fixture()
def molecule_with_bogus_atom():
    molecule = _SimpleMolecule()
    molecule.add_atom(atomic_number=6)
    molecule.add_atom(atomic_number=-1)
    molecule.add_bond(0, 1)

    return molecule


@pytest.fixture()
def t4():
    return _SimpleMolecule.from_molecule(Topology.from_pdb(get_data_file_path("proteins/T4-protein.pdb")).molecule(0))


class TestMMMolecule:
    def test_create_water(self):
        water = _SimpleMolecule()
        water.add_atom(atomic_number=8)
        water.add_atom(atomic_number=1)
        water.add_atom(atomic_number=1)
        water.add_bond(0, 1)
        water.add_bond(0, 2)
        assert water.n_atoms == 3
        assert water.n_bonds == 2
        assert water.n_conformers == 0

    def test_bond_getters(self, water):
        assert water.get_bond_between(0, 1) is not water.get_bond_between(0, 2)
        assert water.bond(0) is not water.bond(1)

        assert water.get_bond_between(0, 1) is water.bond(0)
        assert water.get_bond_between(0, 2) is water.bond(1)

    def test_hill_formula(
        self,
        water,
        molecule_with_zero_atom,
        molecule_with_bogus_atom,
    ):
        assert water.hill_formula == "H2O"
        assert molecule_with_zero_atom.hill_formula == "CX"
        assert molecule_with_bogus_atom.hill_formula == "INVALID"

    def test_to_networkx(self, water):
        graph = water.to_networkx()

        assert graph.number_of_nodes() == water.n_atoms
        assert graph.number_of_edges() == water.n_bonds

    def test_add_conformer(self, water):
        water_molecule = Molecule.from_smiles("O")
        water_molecule.generate_conformers(n_conformers=1)

        assert water.n_conformers == 0
        water.add_conformer(water_molecule.conformers[0])
        assert water.n_conformers == 1

        assert np.allclose(water.conformers[0], water_molecule.conformers[0])

    def test_dict_roundtrip(self, water):
        roundtrip = _SimpleMolecule.from_dict(water.to_dict())

        assert roundtrip.n_atoms == water.n_atoms
        assert roundtrip.n_bonds == water.n_bonds

        for atom_index in range(roundtrip.n_atoms):
            assert roundtrip.atom(atom_index).atomic_number == water.atom(atom_index).atomic_number

    def test_dict_roundtrip_conformers(self, water):
        water_molecule = Molecule.from_smiles("O")
        water_molecule.generate_conformers(n_conformers=1)

        water.add_conformer(water_molecule.conformers[0])

        roundtrip = _SimpleMolecule.from_dict(water.to_dict())

        assert water.n_conformers == roundtrip.n_conformers
        assert np.allclose(water.conformers[0], roundtrip.conformers[0])

    def test_from_molecule(self):
        converted = _SimpleMolecule.from_molecule(Molecule.from_smiles("O"))

        assert converted.n_atoms == 3
        assert converted.n_bonds == 2

        expected_atomic_numbers = [8, 1, 1]
        for atom_index in range(converted.n_atoms):
            found = converted.atom(atom_index).atomic_number
            assert found == expected_atomic_numbers[atom_index]

    def test_deepcopy(self, water, methane, methanol):
        for molecule in [water, methane, methanol]:
            molecule_copy = copy.deepcopy(molecule)

            assert molecule_copy.n_atoms == molecule.n_atoms
            assert molecule_copy.n_bonds == molecule.n_bonds

            for atom, atom_copy in zip(
                molecule.atoms,
                molecule_copy.atoms,
            ):
                assert atom.atomic_number == atom_copy.atomic_number

                assert atom.molecule is molecule
                assert atom_copy.molecule is molecule_copy

    @pytest.mark.slow
    @requires_rdkit
    def test_deepcopy_t4(self, t4):
        t4_copy = copy.deepcopy(t4)

        assert t4_copy.n_atoms == t4.n_atoms
        assert t4_copy.n_bonds == t4.n_bonds

        for atom, atom_copy in zip(
            t4.atoms,
            t4_copy.atoms,
        ):
            assert atom.atomic_number == atom_copy.atomic_number

            assert atom.molecule is t4
            assert atom_copy.molecule is t4_copy

    def test_to_single_molecule_topology(self, methanol):
        topology = methanol.to_topology()

        assert topology.n_molecules == 1
        assert topology.molecule(0).is_isomorphic_with(methanol)

        assert topology.n_atoms == methanol.n_atoms
        assert topology.n_bonds == methanol.n_bonds

    @pytest.mark.slow
    @requires_rdkit
    def test_to_t4_topology(self, t4):
        topology = t4.to_topology()

        assert topology.n_molecules == 1
        assert topology.molecule(0).is_isomorphic_with(t4)

        assert topology.n_atoms == t4.n_atoms
        assert topology.n_bonds == t4.n_bonds

    def test_to_openmm_topology(self, methanol):
        topology = methanol.to_topology().to_openmm()

        assert topology.getNumAtoms() == methanol.n_atoms
        assert topology.getNumBonds() == methanol.n_bonds

    @pytest.mark.slow
    @pytest.mark.skip(reason="Fails because of https://github.com/openforcefield/openff-toolkit/issues/1783")
    def test_to_openmm_topology_t4(self, t4):
        topology = t4.to_topology().to_openmm()

        assert topology.getNumAtoms() == t4.n_atoms
        assert topology.getNumBonds() == t4.n_bonds

    def test_generate_unique_atom_names(self, water):
        # Initially atom names default to empty string
        assert hasattr(water.atom(0), "name")
        assert water.atom(0).name == ""

        assert water.has_unique_atom_names is False
        # Assign unique atom names and ensure that they're unique
        water.generate_unique_atom_names()
        assert water.has_unique_atom_names is True
        # Now manually make them not-unique and ensure that has_unique_atom_names reflects that
        water.atom(0).name = "foo"
        water.atom(1).name = "foo"
        assert water.has_unique_atom_names is False

    def test_atom_names_roundtrip(self, water):
        water.atom(0).name = "FOO"
        water.atom(1).name = "BAR"
        water.atom(2).name = "BAZ"

        roundtrip = copy.deepcopy(water)

        assert roundtrip.atom(0).name == "FOO"
        assert roundtrip.atom(1).name == "BAR"
        assert roundtrip.atom(2).name == "BAZ"


class TestSimpleAtom:
    def test_atom_name_in_dict(self):
        atom = _SimpleAtom(atomic_number=7, name="FLAG")

        assert atom.to_dict()["name"] == "FLAG"

        assert copy.deepcopy(atom).name == "FLAG"


class TestImpropers:
    @pytest.mark.parametrize(
        ("smiles", "n_impropers", "n_pruned"),
        [
            ("C", 24, 0),
            ("CC", 48, 0),
            ("N", 6, 6),
        ],
    )
    def test_pruned_impropers(self, smiles, n_impropers, n_pruned):
        """See equivalent test in TestMolecule."""
        molecule = _SimpleMolecule.from_molecule(
            Molecule.from_smiles(smiles),
        )

        assert molecule.n_impropers == n_impropers
        assert len(list(molecule.smirnoff_impropers)) == n_pruned
        assert len(list(molecule.amber_impropers)) == n_pruned

        amber_impropers = {*molecule.amber_impropers}

        for smirnoff_imp in molecule.smirnoff_impropers:
            assert (
                smirnoff_imp[1],
                smirnoff_imp[0],
                smirnoff_imp[2],
                smirnoff_imp[3],
            ) in amber_impropers


class TestIsomorphism:
    @pytest.fixture()
    def n_propanol(self):
        return _SimpleMolecule.from_molecule(
            Molecule.from_mapped_smiles("[H:5][C:1]([H:6])([H:7])[C:2]([H:8])([H:9])[C:3]([H:10])([H:11])[O:4][H:12]")
        )

    @pytest.fixture()
    def iso_propanol(self):
        return _SimpleMolecule.from_molecule(
            Molecule.from_mapped_smiles("[H:5][C:1]([H:6])([H:7])[C:2]([H:8])([C:3]([H:10])([H:11])[H:12])[O:4][H:9]")
        )

    @pytest.fixture()
    def o_dichlorobezene(self):
        return _SimpleMolecule.from_molecule(
            Molecule.from_mapped_smiles(
                "[C:1]=1([Cl:7])[C:2]([H:9])=[C:3]([H:10])[C:4]([H:11])=[C:5]([H:12])[C:6]([Cl:8])=1"
            )
        )

    @pytest.fixture()
    def m_dichlorobezene(self):
        return _SimpleMolecule.from_molecule(
            Molecule.from_mapped_smiles(
                "[C:1]([Cl:7])(=[C:2]1([H:9]))[C:3]([H:10])=[C:4]([H:11])[C:5]([H:12])=[C:6]([Cl:8])1"
            )
        )

    @pytest.mark.parametrize("as_graphs", [True, False])
    def test_are_isomorphic(self, water, methane, methanol, as_graphs):
        if as_graphs:
            water = water.to_networkx()
            methane = methane.to_networkx()
            methanol = methanol.to_networkx()

        assert _SimpleMolecule.are_isomorphic(water, water)[0]
        assert _SimpleMolecule.are_isomorphic(methane, methane)[0]
        assert _SimpleMolecule.are_isomorphic(methanol, methanol)[0]
        assert not _SimpleMolecule.are_isomorphic(water, methane)[0]
        assert not _SimpleMolecule.are_isomorphic(water, methanol)[0]
        assert not _SimpleMolecule.are_isomorphic(methane, methanol)[0]

    def test_is_isomorphic_with(self, water, methane, methanol):
        assert water.is_isomorphic_with(water)
        assert methane.is_isomorphic_with(methane)
        assert methanol.is_isomorphic_with(methanol)
        assert not water.is_isomorphic_with(methane)
        assert not water.is_isomorphic_with(methanol)
        assert not methane.is_isomorphic_with(methanol)

    def test_short_circuit_heterogeneous_input(self):
        assert not _SimpleMolecule.are_isomorphic(
            create_ethanol(),
            _SimpleMolecule.from_molecule(create_ethanol()),
        )[0]

        assert not _SimpleMolecule.are_isomorphic(
            _SimpleMolecule.from_molecule(create_ethanol()),
            create_ethanol(),
        )[0]

    def test_graph_and_molecule_inputs(self, methanol):
        graph = methanol.to_networkx()

        assert _SimpleMolecule.are_isomorphic(methanol, graph)[0]
        assert _SimpleMolecule.are_isomorphic(graph, methanol)[0]

    def test_propanol_isopropanol_not_isomorphic(self, n_propanol, iso_propanol):
        assert not _SimpleMolecule.are_isomorphic(n_propanol, iso_propanol)[0]

    def test_o_m_dichlorobenzene_not_isomorphic(self, o_dichlorobezene, m_dichlorobezene):
        assert not _SimpleMolecule.are_isomorphic(o_dichlorobezene, m_dichlorobezene)[0]


class TestHierarchyData:
    @pytest.fixture()
    def simple_dipeptide(self):
        return _SimpleMolecule.from_molecule(create_dipeptide())

    def test_mm_hierarchy(self):
        molecule = create_dipeptide()

        assert molecule.hierarchy_schemes is not None

        mm_molecule = _SimpleMolecule.from_molecule(molecule)

        assert mm_molecule.hierarchy_schemes.keys() == molecule.hierarchy_schemes.keys()

        for scheme_name in molecule.hierarchy_schemes:
            assert len(mm_molecule.hierarchy_schemes[scheme_name].hierarchy_elements) == len(
                molecule.hierarchy_schemes[scheme_name].hierarchy_elements
            )

    def test_hierarchy_preserved_dict_roundtrip(self, simple_dipeptide):
        roundtripped = _SimpleMolecule.from_dict(simple_dipeptide.to_dict())

        assert [residue.residue_name for residue in roundtripped.residues] == [
            "ACE",
            "ALA",
        ]

        assert [residue.residue_number for residue in roundtripped.residues] == [
            "1",
            "2",
        ]

    def test_lookup_attribute_not_found(self, simple_dipeptide):
        with pytest.raises(
            AttributeError,
            match=r"object has no attribute .*foobars",
        ):
            simple_dipeptide.foobars



# flake8: noqa
"""
Tools for generating library of substructures from the Chemical Component
Dictionary (CCD).
"""

import copy
import json
from collections import OrderedDict, defaultdict
from copy import deepcopy

import rdkit
from CifFile import ReadCif
from openff.units.elements import SYMBOLS
from rdkit import Chem

from openff.toolkit import Molecule
from openff.toolkit.utils.rdkit_wrapper import RDKitToolkitWrapper

_SYMBOL_TO_ATOMIC_NUMBER = {v: k for k, v in SYMBOLS.items()}


# Amber ff porting functions
def remove_charge_and_bond_order_from_guanidinium(rdmol):
    """
    To correct for chemical perception issues with possible resonance states of arginine,
    remove all charge from the guanidinium group, and set all bond orders to 4. This will
    mark the resonant bonds with a unique "$" character in the SMARTS, which we can later
    replace.
    """
    for atom in rdmol.GetAtoms():
        # print(dir(atom))
        if atom.GetAtomicNum() != 6:  # element.symbol != "C":
            continue
        nitrogen_neighbors = 0
        for neighbor in atom.GetNeighbors():
            if neighbor.GetAtomicNum() == 7:
                nitrogen_neighbors += 1
        if nitrogen_neighbors != 3:
            continue
        atom.SetFormalCharge(0)
        for neighbor in atom.GetNeighbors():
            neighbor.SetFormalCharge(0)
        for bond in atom.GetBonds():
            # Set bond order 4, which will produce a "$" character. We later replace this with "~".
            # print(dir(bond))
            bond.SetBondType(rdkit.Chem.BondType.QUADRUPLE)
            # bond.SetBondType(Chem.BondType.AROMATIC)
            # bond.SetBondType(Chem.BondType.OTHER)


def remove_charge_and_bond_order_from_imidazole(rdmol):
    """
    To correct for chemical perception issues with possible resonance states of histidine,
    remove all charge from the imidazole group, and set all bond orders to 4. This will
    mark the resonant bonds with a unique "$" character in the SMARTS, which we can later
    replace.
    """
    # matches = offmol.chemical_environment_matches('[C:1]1~[C:2]~[N:3]~[C:4]~[N:5]1')
    imidazole_substructure = rdkit.Chem.MolFromSmarts("[C:1]1~[C:2]~[N:3]~[C:4]~[N:5]1")
    matches = rdmol.GetSubstructMatches(imidazole_substructure)
    all_imidazole_atoms = set()
    for match in matches:
        for idx in match:
            all_imidazole_atoms.add(idx)

    for atom in rdmol.GetAtoms():
        if atom.GetIdx() in all_imidazole_atoms:
            atom.SetFormalCharge(0)

    for bond in rdmol.GetBonds():
        # print(dir(bond))
        if (bond.GetBeginAtomIdx() in all_imidazole_atoms) and (bond.GetEndAtomIdx() in all_imidazole_atoms):
            bond.SetBondType(rdkit.Chem.BondType.QUADRUPLE)


class CifSubstructures:
    """
    Substructure from CIF formatted files.
    """

    # bond order dictionary translation from mmcif
    bond_order_dict = {"SING": 1, "DOUB": 2, "TRIP": 3, "QUAD": 4}

    def __init__(self):
        """
        Create object with substructures data from CIF files.
        """
        self.data = defaultdict(defaultdict)  # Dictionary where to store substructures data
        self._cif_multi_entry_object = None

    def from_file(
        self,
        cif_file,
        discard_keyword="FRAGMENT",
        replace_quadruple_bond_with_any: bool = True,
        remove_charge_bond_order_resonant: bool = True,
    ):
        """
        Reads cif formatted file and fills substructure information.

        Sequential runs combine information from different files.

        Parameters
        __________
        cif_file : str or file-like object
            File path string, URL or file-like object.
        discard_keyword : str, optional, default='FRAGMENT'
            Keyword in _chem_comp.name for filtering out entries. Default is 'FRAGMENT'.
        remove_charge_bond_order_resonant: bool, optional, default=True
            Removes bond order and charge information from guanidium and imidazole resonant groups.
        replace_quadruple_bond_with_any: bool, optional, default=True:
            Replace expected unique quadruple bond symbol ($) with any order symbol (~) in smarts patterns.
        """
        # read cif file
        self._cif_multi_entry_object = ReadCif(cif_file)
        # fill data dictionary
        self._fill_substructure_data(
            discard_keyword=discard_keyword,
            replace_quadruple_bond_with_any=replace_quadruple_bond_with_any,
            remove_charge_bond_order_resonant=remove_charge_bond_order_resonant,
        )

    def to_json_file(self, output_file, indent=4):
        """
        Return a JSON serialized representation.

        Parameters
        __________
        output_file : str
            Path string for output file.
        indent : int, optional, default=4
            If not None, will pretty-print with specified number of spaces for indentation.

        Returns
        _______
        serialized : str
            A JSON serialized representation of the object.
        """
        with open(output_file, "w") as library_file:
            json.dump(self.data, library_file, indent=indent)

    def clear_data(self):
        """
        Clears current data structures read from files.

        Useful if you want to start reading cif files from scratch.
        """
        self.data = defaultdict(defaultdict)
        self._cif_multi_entry_object = None

    @staticmethod
    def _generate_atom_symbol_number_dictionary(cif_entry):
        """
        Generate symbol to atomic number dictionary, given entry information from cif file.

        Parameters
        __________
        cif_entry : CifFile.StarFile.StarBlock
            Data block entry from cif file.

        Returns
        _______
        symbol_to_num_dict : dict
            Atom symbol to atomic number (key, value) dictionary.
        """
        # Is this de-duplicated? i.e. can it be ["C", "C", "H"] or only ["C", "H"] ?
        atom_symbol_list: list[str] = cif_entry["_chem_comp_atom.type_symbol"]

        # Is this dict a de-duplicated mapping between element symbols and atomic numbers
        # for the elements in this cif file? If so, can maybe do it in one step
        """
        symbol_to_num_dict: dict[str, int] = {
            symbol: atomic_number for atomic_number, symbol in SYMBOLS.items() if atomic_number in
            cif_entry["_chem_comp_atom.type_symbol"]
        }
        """

        symbol_to_num_dict: dict[str, int] = {symbol: _SYMBOL_TO_ATOMIC_NUMBER[symbol] for symbol in atom_symbol_list}

        return symbol_to_num_dict

    @staticmethod
    def _get_atom_by_name(molecule, name: str):
        """
        Gets first occurrence of atom object from a molecule by its name.

        Parameters
        __________
        molecule : Molecule
            Molecule object where to look.
        name : str
            Atom name string to match against molecule atoms.

        Returns
        _______
        atom_match : Atom or None
            First occurrence of Atom with the given name, in molecule. None if not found.
        """
        for atom in molecule.atoms:
            if atom.name == name:
                match = atom
                break
        else:
            match = None
        return match

    def _gather_atoms_information(self, cif_entry_data):
        """
        Gather minimum required atom information for creating toolkit Atoms.

        Returns
        _______
        atom_information_array : tuple
            Multidimensional tuple with atoms information.
        """
        symbol_to_num_dict = self._generate_atom_symbol_number_dictionary(cif_entry_data)
        atom_names = cif_entry_data["_chem_comp_atom.atom_id"]
        atomic_numbers = [symbol_to_num_dict[x] for x in cif_entry_data["_chem_comp_atom.type_symbol"]]
        formal_charges = [int(x) for x in cif_entry_data["_chem_comp_atom.charge"]]
        is_aromatic = [False if x == "N" else True for x in cif_entry_data["_chem_comp_atom.pdbx_aromatic_flag"]]
        stereochemistry = [None if x == "N" else x for x in cif_entry_data["_chem_comp_atom.pdbx_stereo_config"]]
        leaving_atoms_list = cif_entry_data["_chem_comp_atom.pdbx_leaving_atom_flag"]
        return (
            atom_names,
            atomic_numbers,
            formal_charges,
            is_aromatic,
            stereochemistry,
            leaving_atoms_list,
        )

    def _gather_bonds_information(self, cif_entry_data):
        """
        Gather minimum required bond information for creating toolkit Bonds.

        Returns
        _______
        bond_information_array : tuple
            Multidimensional tuple with bonds information.
        """
        atom1_name_list = cif_entry_data["_chem_comp_bond.atom_id_1"]
        atom2_name_list = cif_entry_data["_chem_comp_bond.atom_id_2"]
        bond_order_list = [self.bond_order_dict[x] for x in cif_entry_data["_chem_comp_bond.value_order"]]
        is_aromatic_bond_list = [
            False if x == "N" else True for x in cif_entry_data["_chem_comp_bond.pdbx_aromatic_flag"]
        ]
        stereochemistry_bond_list = [
            None if x == "N" else x for x in cif_entry_data["_chem_comp_bond.pdbx_stereo_config"]
        ]
        return (
            atom1_name_list,
            atom2_name_list,
            bond_order_list,
            is_aromatic_bond_list,
            stereochemistry_bond_list,
        )

    @staticmethod
    def _get_subrdmol(
        rdmol: rdkit.Chem.rdchem.Mol,
        indices: list = [],
        sanitize: bool = False,
        remove_charge_bond_order_resonant: bool = True,
    ):
        """
        Create new sub-molecule from selected atom indices

        Parameters
        ----------
        rdmol: rdkit.Chem.rdchem.Mol
            Input molecule
        indices: iterable of ints
            atom indices to include from input molecule, indexed from 0
        sanitize: bool
            whether to sanitize the molecule (recommend: no)

        Returns
        -------
        rdkit.Chem.rdchem.Mol: subset of molecule
        """
        submol = rdkit.Chem.RWMol(rdmol)
        ix = sorted([at.GetIdx() for at in rdmol.GetAtoms() if at.GetIdx() not in indices])
        for i in ix[::-1]:
            submol.RemoveAtom(int(i))
        if sanitize:
            rdkit.Chem.SanitizeMol(submol)

        for atom in submol.GetAtoms():
            # print(dir(atom))
            atom.SetNoImplicit(True)

        if remove_charge_bond_order_resonant:
            remove_charge_and_bond_order_from_imidazole(submol)
            remove_charge_and_bond_order_from_guanidinium(submol)
        return submol

    def _get_smiles(
        self,
        mol: Molecule,
        indices: list = [],
        label_indices: list = [],
        replace_quadruple_bond_with_any: bool = True,
        remove_charge_bond_order_resonant: bool = True,
    ):
        """
        Get SMARTS of selected atoms in molecule

        Parameters
        ----------
        mol: openff.toolkit.Molecule
            Input molecule
        indices: iterable of ints
            atom indices to include in SMARTS, indexed from 0
        label_indices: iterable of ints
            atom indices to label, indexed from 0. The atoms
            will be labelled in the order specified. Labels
            begin from 1.

        Returns
        -------
        str: SMARTS string
        """
        rdmol = mol.to_rdkit()
        for i, lix in enumerate(label_indices, 1):
            at = rdmol.GetAtomWithIdx(int(lix))
            at.SetAtomMapNum(i)
        indices = sorted(set(indices) | set(label_indices))
        submol = self._get_subrdmol(
            rdmol,
            indices,
            remove_charge_bond_order_resonant=remove_charge_bond_order_resonant,
        )
        smiles = rdkit.Chem.MolToSmiles(submol, allHsExplicit=True)
        # TODO: This doesn't seem to be having an effect. It is replacing bond orders regardless.
        if replace_quadruple_bond_with_any:
            smiles = smiles.replace("$", "~")
        return smiles

    def _add_substructure_data_entry(
        self,
        atoms_information,
        bonds_information,
        entry_code,
        replace_quadruple_bond_with_any: bool = True,
        remove_charge_bond_order_resonant: bool = True,
    ):
        """
        Read substructure from entry data in cif file and fill data object with smiles.

        Parameters
        __________
        atoms_information : list[tuple[str,int]]
            Iterable of tuple of atom info.
            Currently [atom_name, element, formal_charge, is_aromatic, stereo, is_leaving]
        bonds_information : list[tuple[str,int]]
            Iterable of tuple of bond info.
            Currently [atom1_name, atom2_name, bond_order, is_aromatic, stereo]
        """
        offmol = Molecule()

        atom_name_to_idx = dict()
        # add atoms
        for (
            atom_name,
            atomic_number,
            formal_charge,
            is_aromatic,
            stereochemistry,
            is_leaving,
        ) in atoms_information:
            # if include_leaving:
            # Add all atoms
            atom_idx = offmol.add_atom(
                atomic_number,
                formal_charge,
                is_aromatic,
                stereochemistry=stereochemistry,
                name=atom_name,
            )
            atom_name_to_idx[atom_name] = atom_idx
        for (
            atom1_name,
            atom2_name,
            bond_order,
            is_aromatic,
            stereochemistry,
        ) in bonds_information:
            # try:
            atom1_idx = atom_name_to_idx[atom1_name]
            atom2_idx = atom_name_to_idx[atom2_name]
            offmol.add_bond(
                atom1_idx,
                atom2_idx,
                bond_order,
                is_aromatic=is_aromatic,
                stereochemistry=stereochemistry,
            )
            # # TODO: Broad exception. We should raise a specific error.
            # except Exception:  # happens when atom is not found, i.e. no leaving atoms.
            #     continue
        # Get smiles
        atom_indices = list(range(offmol.n_atoms))
        smiles = self._get_smiles(
            offmol,
            indices=atom_indices,
            label_indices=atom_indices,
            replace_quadruple_bond_with_any=replace_quadruple_bond_with_any,
            remove_charge_bond_order_resonant=remove_charge_bond_order_resonant,
        )
        if smiles not in self.data[entry_code]:
            atom_names = [atom_info[0] for atom_info in atoms_information]
            self.data[entry_code][smiles] = atom_names

    def _fill_substructure_data(
        self,
        discard_keyword="FRAGMENT",
        replace_quadruple_bond_with_any: bool = True,
        remove_charge_bond_order_resonant: bool = True,
    ):
        """
        Fills data dictionary with the substructure information.

        Parameters
        __________
        discard_keyword : str, optional, default='FRAGMENT'
            Keyword in _chem_comp.name for filtering out entries. Default is 'FRAGMENT'.
        """
        # override_dict contains atoms that we DO want to treat as leaving, even
        # though the actual entries don't have them flagged as leaving. The "H2" entries
        # in this list are hydrogens attached to the backbone N, which are (for some reason)
        # recorded as NH2, with neither H marked as "leaving", which causes the
        # substructures to be unable to match main-chain appearances of the amino acids.
        override_dict = {
            # LYN
            "lys_lfoh_dhz3": ["H2"],
            # GLU
            "glu_lfoh_dhe2": ["H2"],
            # HIE
            "his_lfoh_dhd1": ["H2"],
            # ASP
            "asp_lfoh_dhd2": ["H2"],
            # HID
            "his_lfoh_dhe2": ["H2"],
            # CYX
            "cys": ["HG"],
            # CYM
            "cys_lfoh_dhg": ["H1", "H2"],
        }

        for entry, entry_data in self._cif_multi_entry_object.items():
            # Skip entries matching discard filtering keyword in comp name
            comp_name = entry_data["_chem_comp.name"]
            if discard_keyword in comp_name:
                continue
            # Gather necessary data for creating/adding atoms
            atoms_information = [*zip(*self._gather_atoms_information(entry_data))]
            # Gather information for bonds
            bonds_information = [*zip(*self._gather_bonds_information(entry_data))]

            for override_list in override_dict.get(entry, []):
                for idx, atom in enumerate(atoms_information):
                    if atom[0] in override_list:
                        new_atom_tuple = (
                            atom[0],
                            atom[1],
                            atom[2],
                            atom[3],
                            atom[4],
                            "Y",
                        )
                        atoms_information[idx] = new_atom_tuple
            # Only take three letter code for key -- Note it uses the new atom_names list
            entry_code = entry_data["_chem_comp.three_letter_code"]
            # Generate structures with permutations of leaving atoms. For example, if a COH
            # motif has -OH group marked as leaving, this will add
            # 1) the O- intermediate where only the H has left, as well as
            # 2) the structure with the entire OH removed.
            # Only atoms marked as "leaving" may have their formal charge changed by
            # this function, so the remaining C when the OH is removed still has a
            # neutral charge.
            info_tuples = self._recursive_prepare_atom_bond_info(
                atoms_information, bonds_information, return_list=list()
            )

            # Also add a structure with ALL leaving atoms removed
            atoms_info_copy = copy.deepcopy(atoms_information)
            bonds_info_copy = copy.deepcopy(bonds_information)
            atom_idxs_to_remove = [idx for idx, at in enumerate(atoms_info_copy) if at[5] == "Y"]
            atom_names_to_remove = [atoms_info_copy[idx][0] for idx in atom_idxs_to_remove]
            bond_idxs_to_remove = list()
            for bond_idx, bond in enumerate(bonds_info_copy):
                if (bond[0] in atom_names_to_remove) or (bond[1] in atom_names_to_remove):
                    bond_idxs_to_remove.append(bond_idx)

            for atom_idx_to_remove in atom_idxs_to_remove[::-1]:
                atoms_info_copy.pop(atom_idx_to_remove)
            for bond_idx_to_remove in bond_idxs_to_remove[::-1]:
                bonds_info_copy.pop(bond_idx_to_remove)
            info_tuples.append((atoms_info_copy, bonds_info_copy))

            # Generate substructure entries for all prepared substructures
            for atoms_information, bonds_information in info_tuples:
                self._add_substructure_data_entry(
                    atoms_information,
                    bonds_information,
                    entry_code,
                    replace_quadruple_bond_with_any=replace_quadruple_bond_with_any,
                    remove_charge_bond_order_resonant=remove_charge_bond_order_resonant,
                )

    def _recursive_prepare_atom_bond_info(self, atoms_info, bonds_info, return_list=list()):
        """
        This method makes permutations of the substructure with leaving atoms left and removed.
        If some leaving atoms are bound to other leaving atoms, this method produces all permutations
        of the leaving atom chains that are possible without creating a disconnected molecular graph.

        If multiple leaving atoms form a chain, then this method attempts to keep the graph chemically
        valid by subtracting one from the formal charge of each leaving atom when its neighbor is removed.
        However non-leaving atoms will never have their formal charge changed.

        Parameters
        ----------
        atoms_info : list[tuple[str,int]]
            Iterable of tuple of atom info.
            Currently [atom_name, element, formal_charge, is_aromatic, stereo, is_leaving]
        bonds_info : list[tuple[str,int]]
            Iterable of tuple of bond info.
            Currently [atom1_name, atom2_name, bond_order, is_aromatic, stereo]
        """
        return_list.append((atoms_info, bonds_info))
        for leaving_atom_idx, leaving_atom_info_tuple in enumerate(atoms_info):
            # If the atom isn't leaving then we don't need to do anything
            if leaving_atom_info_tuple[5] == "N":
                continue
            # If the atom IS leaving, then find any bonds it's involved in
            leaving_atom_name = leaving_atom_info_tuple[0]
            involved_bond_indices = list()
            for bond_idx, bond_info_tuple in enumerate(bonds_info):
                if leaving_atom_name in bond_info_tuple[:2]:
                    involved_bond_indices.append(bond_idx)
            # If this leaving atom is involved in more than one bond, then
            # don't remove it, since doing so would disconnect the molecule
            if len(involved_bond_indices) != 1:
                continue
            atoms_info_copy = copy.deepcopy(atoms_info)
            bonds_info_copy = copy.deepcopy(bonds_info)

            # Determine whether the neighbor of the leaving atom is also leaving
            bond_info_tuple = bonds_info[involved_bond_indices[0]]
            neighbor_atom_name = [name for name in bond_info_tuple[:2] if name != leaving_atom_name][0]
            for neighbor_atom_idx, neighbor_atom_info_tuple in enumerate(atoms_info):
                if neighbor_atom_info_tuple[0] != neighbor_atom_name:
                    continue
                neighbor_atom_is_also_leaving = neighbor_atom_info_tuple[5] == "Y"
                # If the neighbor atom is ALSO leaving, then deduct 1 from its formal charge
                if neighbor_atom_is_also_leaving:
                    new_atom_tuple = (
                        neighbor_atom_info_tuple[0],
                        neighbor_atom_info_tuple[1],
                        neighbor_atom_info_tuple[2] - 1,
                        neighbor_atom_info_tuple[3],
                        neighbor_atom_info_tuple[4],
                        neighbor_atom_info_tuple[5],
                    )
                    atoms_info_copy[neighbor_atom_idx] = new_atom_tuple
            atoms_info_copy.pop(leaving_atom_idx)
            bonds_info_copy.pop(involved_bond_indices[0])
            return_list = self._recursive_prepare_atom_bond_info(
                atoms_info_copy, bonds_info_copy, return_list=return_list
            )

        return return_list

    def _patch_known_problems(self):
        """
        Monkey-patching known problems with current aa-variants-v1.cif from CCD.

        .. warning: Needed as of Oct-21-2021
        """

        substructures_to_fix = {
            "PRO": [
                (
                    "[N-:1]1[C@@:2]([C-:3]=[O:4])([H:8])[C:5]([H:9])([H:10])[C:6]([H:11])([H:12])[C:7]1([H:13])[H:14]",
                    "[N:1]1[C@@:2]([C:3]=[O:4])([H:8])[C:5]([H:9])([H:10])[C:6]([H:11])([H:12])[C:7]1([H:13])[H:14]",
                ),
            ],
            "HIS": [
                (
                    "[N:1]([C@:2]([C:3](=[O:4])[O:11][H:18])([C:5]([C:6]1=[C:8]([H:16])[N-:10][C:9]([H:17])=[N+:7]1[H:15])([H:13])[H:14])[H:12])([H:19])[H:20]",
                    "[N:1]([C@:2]([C:3](=[O:4])[O:11][H:18])([C:5]([C:6]1=[C:8]([H:16])[N:10]=[C:9]([H:17])[N:7]1[H:15])([H:13])[H:14])[H:12])([H:19])[H:20]",
                ),
                (
                    "[N:1]([C@:2]([C:3](=[O:4])[O-:11])([C:5]([C:6]1=[C:8]([H:16])[N-:10][C:9]([H:17])=[N+:7]1[H:15])([H:13])[H:14])[H:12])([H:18])[H:19]",
                    "[N:1]([C@:2]([C:3](=[O:4])[O-:11])([C:5]([C:6]1=[C:8]([H:16])[N:10]=[C:9]([H:17])[N:7]1[H:15])([H:13])[H:14])[H:12])([H:18])[H:19]",
                ),
                (
                    "[N:1]([C@:2]([C:3]=[O:4])([C:5]([C:6]1=[C:8]([H:15])[N-:10][C:9]([H:16])=[N+:7]1[H:14])([H:12])[H:13])[H:11])([H:17])[H:18]",
                    "[N:1]([C@:2]([C:3]=[O:4])([C:5]([C:6]1=[C:8]([H:15])[N:10]=[C:9]([H:16])[N:7]1[H:14])([H:12])[H:13])[H:11])([H:17])[H:18]",
                ),
                (
                    "[N:1]([C@:2]([C:3]=[O:4])([C:5]([C:6]1=[C:8]([H:15])[N-:10][C:9]([H:16])=[N+:7]1[H:14])([H:12])[H:13])[H:11])[H:17]",
                    "[N:1]([C@:2]([C:3]=[O:4])([C:5]([C:6]1=[C:8]([H:15])[N:10]=[C:9]([H:16])[N:7]1[H:14])([H:12])[H:13])[H:11])[H:17]",
                ),
                (
                    "[N:1]([C@:2]([C:3](=[O:4])[O-:11])([C:5]([C:6]1=[C:8]([H:16])[N-:10][C:9]([H:17])=[N+:7]1[H:15])([H:13])[H:14])[H:12])[H:18]",
                    "[N:1]([C@:2]([C:3](=[O:4])[O-:11])([C:5]([C:6]1=[C:8]([H:16])[N:10]=[C:9]([H:17])[N:7]1[H:15])([H:13])[H:14])[H:12])[H:18]",
                ),
                (
                    "[N:1]([C@:2]([C:3](=[O:4])[O:11][H:18])([C:5]([C:6]1=[C:8]([H:16])[N-:10][C:9]([H:17])=[N+:7]1[H:15])([H:13])[H:14])[H:12])[H:19]",
                    "[N:1]([C@:2]([C:3](=[O:4])[O:11][H:18])([C:5]([C:6]1=[C:8]([H:16])[N:10]=[C:9]([H:17])[N:7]1[H:15])([H:13])[H:14])[H:12])[H:19]",
                ),
                (
                    "[N+:1]([C@:2]([C:3](=[O:4])[O-:11])([C:5]([C:6]1=[C:8]([H:16])[N-:10][C:9]([H:17])=[N+:7]1[H:15])([H:13])[H:14])[H:12])([H:18])([H:19])[H:20]",
                    "[N+:1]([C@:2]([C:3](=[O:4])[O-:11])([C:5]([C:6]1=[C:8]([H:16])[N:10]=[C:9]([H:17])[N:7]1[H:15])([H:13])[H:14])[H:12])([H:18])([H:19])[H:20]",
                ),
                (
                    "[N+:1]([C@:2]([C:3]=[O:4])([C:5]([C:6]1=[C:8]([H:15])[N-:10][C:9]([H:16])=[N+:7]1[H:14])([H:12])[H:13])[H:11])([H:17])([H:18])[H:19]",
                    "[N+:1]([C@:2]([C:3]=[O:4])([C:5]([C:6]1=[C:8]([H:15])[N:10]=[C:9]([H:16])[N:7]1[H:14])([H:12])[H:13])[H:11])([H:17])([H:18])[H:19]",
                ),
            ],
            "TRP": [
                (
                    "[N:1]([C@:2]([C:3](=[O:4])[O:15][H:24])([C:5]([C:6]1=[C:7]([H:19])[N-:9][c:10]2[c:8]1[c:11]([H:20])[c:13]([H:22])[c:14]([H:23])[c:12]2[H:21])([H:17])[H:18])[H:16])([H:25])[H:26]",
                    "[N:1]([C@:2]([C:3](=[O:4])[O:15][H:24])([C:5]([C:6]1=[C:7]([H:19])[N:9][c:10]2[c:8]1[c:11]([H:20])[c:13]([H:22])[c:14]([H:23])[c:12]2[H:21])([H:17])[H:18])[H:16])([H:25])[H:26]",
                ),
                (
                    "[N:1]([C@:2]([C:3](=[O:4])[O-:15])([C:5]([C:6]1=[C:7]([H:19])[N-:9][c:10]2[c:8]1[c:11]([H:20])[c:13]([H:22])[c:14]([H:23])[c:12]2[H:21])([H:17])[H:18])[H:16])([H:24])[H:25]",
                    "[N:1]([C@:2]([C:3](=[O:4])[O-:15])([C:5]([C:6]1=[C:7]([H:19])[N:9][c:10]2[c:8]1[c:11]([H:20])[c:13]([H:22])[c:14]([H:23])[c:12]2[H:21])([H:17])[H:18])[H:16])([H:24])[H:25]",
                ),
                (
                    "[N:1]([C@:2]([C:3]=[O:4])([C:5]([C:6]1=[C:7]([H:18])[N-:9][c:10]2[c:8]1[c:11]([H:19])[c:13]([H:21])[c:14]([H:22])[c:12]2[H:20])([H:16])[H:17])[H:15])([H:23])[H:24]",
                    "[N:1]([C@:2]([C:3]=[O:4])([C:5]([C:6]1=[C:7]([H:18])[N:9][c:10]2[c:8]1[c:11]([H:19])[c:13]([H:21])[c:14]([H:22])[c:12]2[H:20])([H:16])[H:17])[H:15])([H:23])[H:24]",
                ),
                (
                    "[N+:1]([C@:2]([C:3](=[O:4])[O-:15])([C:5]([C:6]1=[C:7]([H:19])[N-:9][c:10]2[c:8]1[c:11]([H:20])[c:13]([H:22])[c:14]([H:23])[c:12]2[H:21])([H:17])[H:18])[H:16])([H:24])([H:25])[H:26]",
                    "[N+:1]([C@:2]([C:3](=[O:4])[O-:15])([C:5]([C:6]1=[C:7]([H:19])[N:9][c:10]2[c:8]1[c:11]([H:20])[c:13]([H:22])[c:14]([H:23])[c:12]2[H:21])([H:17])[H:18])[H:16])([H:24])([H:25])[H:26]",
                ),
                (
                    "[N+:1]([C@:2]([C:3]=[O:4])([C:5]([C:6]1=[C:7]([H:18])[N-:9][c:10]2[c:8]1[c:11]([H:19])[c:13]([H:21])[c:14]([H:22])[c:12]2[H:20])([H:16])[H:17])[H:15])([H:23])([H:24])[H:25]",
                    "[N+:1]([C@:2]([C:3]=[O:4])([C:5]([C:6]1=[C:7]([H:18])[N:9][c:10]2[c:8]1[c:11]([H:19])[c:13]([H:21])[c:14]([H:22])[c:12]2[H:20])([H:16])[H:17])[H:15])([H:23])([H:24])[H:25]",
                ),
            ],
        }
        # Fix PRO smarts substructure
        for aa_name, replacement_list in substructures_to_fix.items():
            for old_smarts, new_smarts in replacement_list:
                self.data[aa_name][new_smarts] = self.data[aa_name][old_smarts]
                self.data[aa_name].pop(old_smarts)

    def _add_common_substructures(self):
        """
        .. warning: Needed as of Oct-21-2021
        """

        # Add common caps
        self.data["ACE"] = {"[C:1](=[O:2])[C:3]([H:4])([H:5])[H:6]": ["C", "O", "CH3", "H1", "H2", "H3"]}
        self.data["NME"] = {"[N:1]([C:2]([H:4])([H:5])[H:6])[H:3]": ["N", "C", "H", "H1", "H2", "H3"]}
        self.data["NH2"] = {"[N:1]([H:2])[H:3]": ["N", "HN1", "HN2"]}

    def _add_common_linkages(self):
        """
        .. warning: Needed as of Oct-21-2021
        """

        self.data["PEPTIDE_BOND"] = {
            "[C:1](=[O:2])[N:3]": [
                "C",
                "O",
                "N",
            ]
        }
        self.data["DISULFIDE"] = {
            "[S:1][S:2]": [
                "SG",
                "SG",
            ]
        }

    def _reformat_substructs(self):
        """
        Implemented 7/7/23: add-on to existing substructure library to reformat amino acids
            to a new format. Self contained for now to the following function:
        """

        def add_atom(
            rdmol,
            begin_atom_idx,
            new_atom_smarts,
            bond_order=Chem.rdchem.BondType.SINGLE,
        ):
            # adds atom with single bond, returns new mol
            emol = Chem.RWMol(deepcopy(rdmol))
            new_atom = Chem.AtomFromSmarts(new_atom_smarts)
            idx = emol.AddAtom(new_atom)
            assert idx == rdmol.GetNumAtoms()
            emol.AddBond(begin_atom_idx, idx, order=bond_order)
            return emol.GetMol()

        def _get_maximum_map_num(rdmol):
            return max([atom.GetAtomMapNum() for atom in rdmol.GetAtoms()])

        def _fill_out_query(rdmol):
            params = Chem.SmilesParserParams()
            params.removeHs = False
            params.sanitize = True
            rdmol = Chem.MolFromSmiles(
                Chem.MolToSmiles(
                    deepcopy(rdmol),
                    isomericSmiles=False,
                    kekuleSmiles=True,
                    canonical=False,
                ),
                params,
            )  # create non-query version of substruct
            Chem.Kekulize(rdmol, clearAromaticFlags=True)  # kekulize using molops (regular molecule required here)
            rdmol = Chem.MolFromSmarts(
                Chem.MolToSmiles(
                    deepcopy(rdmol),
                    isomericSmiles=False,
                    kekuleSmiles=True,
                    canonical=False,
                )
            )  # remake into query mol
            current_map_num = 1 + _get_maximum_map_num(rdmol)  # do not change existing atom map numbers!
            for atom in rdmol.GetAtoms():
                if atom.GetAtomicNum() > 0:
                    if atom.GetAtomMapNum() == 0:
                        atom.SetAtomMapNum(current_map_num)
                        current_map_num += 1
                    atom.SetChiralTag(Chem.rdchem.ChiralType.CHI_UNSPECIFIED)
                    a_num = atom.GetAtomicNum()
                    D_num = len([0 for _ in atom.GetBonds()])
                    F_num = atom.GetFormalCharge()
                    query_string = f"[#{a_num}D{D_num}{F_num:+}:{current_map_num}]"
                    query = Chem.AtomFromSmarts(query_string)
                    atom.SetQuery(query)
                elif atom.GetAtomicNum() == 0:
                    atom.SetAtomMapNum(current_map_num)
                    current_map_num += 1
                    query_string = f"[*:{current_map_num}]"
                    query = Chem.AtomFromSmarts(query_string)
                    atom.SetQuery(query)
                else:
                    raise Exception
            for bond in rdmol.GetBonds():
                if bond.GetBondType() == Chem.rdchem.BondType.AROMATIC:
                    raise Exception  # there should be no aromatic bonds unless kekulization has failed
                elif bond.GetBondType() == Chem.rdchem.BondType.SINGLE:
                    query = Chem.BondFromSmarts("-")
                    bond.SetQuery(query)

            return rdmol

        def get_atom_with_map_num(rdmol, map_num):
            for atom in rdmol.GetAtoms():
                if atom.GetAtomMapNum() == map_num:
                    return atom
            return None

        def clear_query(rdmol):
            for atom in rdmol.GetAtoms():
                atom.SetAtomMapNum(0)
                n = atom.GetAtomicNum()
                c = atom.GetFormalCharge()
                query_string = f"[#{n}{c:+}]"
                query = Chem.AtomFromSmarts(query_string)
                atom.SetQuery(query)

        new_subs_dict = OrderedDict()
        #               = [(Smarts without terminal groups   , [ids to add port to],       [ids to add H to],            [id of amide bond carbon])]
        amino_acid_subs = [
            (
                Chem.MolFromSmarts(
                    "[N+:1](-[H])(-[H])1[C@@:2]([C:3](=[O:4]))([H:9])[C:5]([H:10])([H:11])[C:6]([H:12])([H:13])[C:7]1([H:14])[H:15]"
                ),
                [],
                [],
                4,
            ),
            (
                Chem.MolFromSmarts(
                    "[N+0:1](-[H])1[C@@:2]([C:3](=[O:4]))([H:9])[C:5]([H:10])([H:11])[C:6]([H:12])([H:13])[C:7]1([H:14])[H:15]"
                ),
                [],
                [],
                3,
            ),
            (
                Chem.MolFromSmarts(
                    "[N+1:1](-[H])1[C@@:2]([C:3](=[O:4]))([H:9])[C:5]([H:10])([H:11])[C:6]([H:12])([H:13])[C:7]1([H:14])[H:15]"
                ),
                [0],
                [],
                3,
            ),
            (
                Chem.MolFromSmarts(
                    "[N+0:1]1[C@@:2]([C:3](=[O:4]))([H:9])[C:5]([H:10])([H:11])[C:6]([H:12])([H:13])[C:7]1([H:14])[H:15]"
                ),
                [0],
                [],
                2,
            ),
            (
                Chem.MolFromSmarts(
                    "[N+](-[H])(-[H])(-[H])-[C@:2]([C:3](=[O:4]))([C:5]([S+0:6]([H]))([H:9])[H:10])[H:8]"
                ),
                [],
                [],
                5,
            ),
            (
                Chem.MolFromSmarts("[N+](-[H])(-[H])(-[H])-[C@:2]([C:3](=[O:4]))([C:5]([S+0:6])([H:9])[H:10])[H:8]"),
                [8],
                [],
                5,
            ),
            (
                Chem.MolFromSmarts("[N+](-[H])(-[H])(-[H])-[C@:2]([C:3](=[O:4]))([C:5]([S-1:6])([H:9])[H:10])[H:8]"),
                [],
                [],
                5,
            ),
            (
                Chem.MolFromSmarts("[N](-[H])(-[H])-[C@:2]([C:3](=[O:4]))([C:5]([S+0:6]([H]))([H:9])[H:10])[H:8]"),
                [],
                [],
                4,
            ),
            (
                Chem.MolFromSmarts("[N](-[H])(-[H])-[C@:2]([C:3](=[O:4]))([C:5]([S+0:6])([H:9])[H:10])[H:8]"),
                [7],
                [],
                4,
            ),
            (
                Chem.MolFromSmarts("[N](-[H])(-[H])-[C@:2]([C:3](=[O:4]))([C:5]([S-1:6])([H:9])[H:10])[H:8]"),
                [],
                [],
                4,
            ),
            (
                Chem.MolFromSmarts("[N](-[H])-[C@:2]([C:3](=[O:4]))([C:5]([S+0:6]([H]))([H:9])[H:10])[H:8]"),
                [0],
                [],
                3,
            ),
            (
                Chem.MolFromSmarts("[N](-[H])-[C@:2]([C:3](=[O:4]))([C:5]([S+0:6])([H:9])[H:10])[H:8]"),
                [0, 6],
                [],
                3,
            ),
            (
                Chem.MolFromSmarts("[N](-[H])-[C@:2]([C:3](=[O:4]))([C:5]([S-1:6])([H:9])[H:10])[H:8]"),
                [0],
                [],
                3,
            ),
            (
                Chem.MolFromSmarts("[N:1][C@:2]([C:3](=[O:4]))([C:5]([S+0:6]([H]))([H:9])[H:10])[H:8]"),
                [0, 0],
                [],
                2,
            ),
            (
                Chem.MolFromSmarts("[N:1][C@:2]([C:3](=[O:4]))([C:5]([S+0:6])([H:9])[H:10])[H:8]"),
                [0, 0, 5],
                [],
                2,
            ),
            (
                Chem.MolFromSmarts("[N:1][C@:2]([C:3](=[O:4]))([C:5]([S-1:6])([H:9])[H:10])[H:8]"),
                [0, 0],
                [],
                2,
            ),
            (
                Chem.MolFromSmarts("[N+](-[H])(-[H])(-[H])-[C@](-[H])(-[*])-[C](=[O])"),
                [],
                [],
                7,
            ),
            (
                Chem.MolFromSmarts("[N](-[H])(-[H])-[C@](-[H])(-[*])-[C](=[O])"),
                [],
                [],
                6,
            ),
            (Chem.MolFromSmarts("[N](-[H])-[C@](-[H])(-[*])-[C](=[O])"), [0], [], 5),
        ]
        special_cases = {
            "[C:1](=[O:2])[C:3]([H:4])([H:5])[H:6]": "[*]-[C:1](=[O:2])[C:3]([H:4])([H:5])[H:6]",
            "[N:1]([C:2]([H:4])([H:5])[H:6])[H:3]": "[*]-[N:1]([C:2]([H:4])([H:5])[H:6])[H:3]",
            "[N:1]([H:2])[H:3]": "[*]-C(=O)[N:1]([H:2])[H:3]",
            "[C:1](=[O:2])[N:3]": "[*]-[C:1](=[O:2])[N:3](-[H])[C:4](-[*])(-[*])(-[*])",
            "[S:1][S:2]": "[*]-[S:1]-[S:2]-[*]",
        }
        for res_name, subs in self.data.items():
            res_dict = OrderedDict()
            res_num = 0
            for substruct, atom_names in subs.items():
                rdmol = Chem.MolFromSmarts(substruct)
                old_rdmol = deepcopy(rdmol)
                query = tuple()
                add_port = []
                add_Hs = []
                carboxyl_C = -1
                for query, add_port, add_Hs, carboxyl_C in amino_acid_subs:
                    aa_match = rdmol.GetSubstructMatch(query)
                    if aa_match:
                        break
                if not aa_match and substruct not in special_cases.keys():
                    raise Exception(f"{res_name}, {substruct} has no match")
                if substruct in special_cases.keys():
                    if special_cases[substruct]:
                        rdmol = Chem.MolFromSmarts(special_cases[substruct])
                    else:  # empty, discard
                        continue
                else:
                    # add * where appropriate
                    for port_idx in add_port:
                        rdmol = add_atom(rdmol, aa_match[port_idx], "[*]")
                    # add H where appropriate
                    for H_idx in add_Hs:
                        rdmol = add_atom(rdmol, aa_match[H_idx], "[H]")
                    # add port to carboxyl atom if needed
                    carb = rdmol.GetAtomWithIdx(aa_match[carboxyl_C])
                    if carb.GetDegree() == 2:
                        rdmol = add_atom(rdmol, aa_match[carboxyl_C], "[*]")
                formated_rdmol = _fill_out_query(rdmol)
                # In this document, the only new atoms I add are *, so edit the atom names
                num_new_atoms = formated_rdmol.GetNumAtoms() - len(atom_names)
                atom_names = atom_names + ["*"] * num_new_atoms
                sub_smarts = Chem.MolToSmarts(formated_rdmol)
                res_num += 1

                sub_smarts = sub_smarts.replace("&", "")
                res_dict[sub_smarts] = atom_names

            new_subs_dict[res_name] = res_dict

        # validate using toolkit validator to check
        check_dict = dict()
        for res_name, subs in new_subs_dict.items():
            res_num = 0
            for substruct, atom_names in subs.items():
                check_dict[f"{res_name}_{res_num}"] = [substruct]
                res_num += 1
        RDKitToolkitWrapper()._validate_custom_substructures(
            custom_substructures=check_dict, forbidden_keys=[]
        )  # errors if any are invalid
        self.data = new_subs_dict



"""
Test that the examples in the repo run without errors.
"""

import pathlib
import re
import subprocess
import tempfile
import textwrap

import pytest

from openff.toolkit._tests.utils import _get_readme_path
from openff.toolkit.utils import RDKIT_AVAILABLE, get_data_file_path, temporary_cd


def run_script_file(file_path):
    """Run through the shell a python script."""
    with tempfile.TemporaryDirectory() as tmp_dir:
        with temporary_cd(tmp_dir):
            cmd = ["python", file_path]
            if "conformer_energies.py" in file_path:
                cmd.append("--filename")
                mol_file = get_data_file_path("molecules/ruxolitinib_conformers.sdf")
                cmd.append(mol_file)
            try:
                subprocess.check_call(cmd)
            except subprocess.CalledProcessError:
                raise Exception(f"Example {file_path} failed")


def run_script_str(script_str):
    """Execute a Python string through the shell in a temporary directory.

    With respect to eval, this has the advantage of catching all import errors.

    """
    with tempfile.TemporaryDirectory() as tmp_dir:
        temp_file_path = (pathlib.Path(tmp_dir) / "temp.py").as_posix()
        # Create temporary python script.
        with open(temp_file_path, "w") as f:
            f.write(script_str)
        # Run the Python script.
        try:
            run_script_file(temp_file_path)
        except Exception as error:
            script_str = textwrap.indent(script_str, "    ")
            raise Exception(f"The following script failed:\n{script_str}") from error


def find_example_scripts() -> list[str]:
    """Find all Python scripts, excluding Jupyter notebooks, in the examples folder.

    Returns
    -------
    example_file_paths : list[str]
        List of full paths to python scripts to execute.
    """
    if "site-packages" in __file__:
        # This test file is being collected from the installed package, which
        # does not provide the examples folder in the same location
        return list()

    examples_dir_path = pathlib.Path(__file__).parents[3] / "examples"

    # Examples that require RDKit
    rdkit_examples = {
        examples_dir_path / "conformer_energies/conformer_energies.py",
    }

    example_file_paths = []
    for example_file_path in examples_dir_path.glob("*/*.py"):
        if not RDKIT_AVAILABLE:
            if example_file_path in rdkit_examples:
                continue
        example_file_paths.append(example_file_path.as_posix())

    return example_file_paths


def find_readme_examples() -> list[str]:
    """Yield the Python scripts in the main README.md file.

    Returns
    -------
    readme_examples : list[str]
        The list of Python scripts included in the README.md files.
    """
    readme_file_path = _get_readme_path()

    if readme_file_path is None:
        return list()

    with open(readme_file_path) as f:
        readme_content = f.read()

    return re.findall("```python(.*?)```", readme_content, flags=re.DOTALL)


@pytest.mark.parametrize("readme_example_str", find_readme_examples())
def test_readme_examples(readme_example_str):
    """Test the example"""
    run_script_str(readme_example_str)


@pytest.mark.parametrize("example_file_path", find_example_scripts())
def test_examples(example_file_path):
    """Test that the example run without errors."""
    run_script_file(example_file_path)



"""
Utility classes and functions to create objects supporting callback registration.

"""

__all__ = [
    "Callbackable",
    "callback_method",
]


import functools
import inspect

from openff.toolkit.utils.exceptions import CallbackRegistrationError


def callback_method(func=None, events=()):
    """Decorator used to mark a method as callbackable.

    This decorator is designed to be used by ``Callbackable`` children
    classes. Usage examples can be found in the documentation of ``Callbackable``.

    Parameters
    ----------
    func : MethodType
        The instance method to decorate. Static and class methods are
        currently not supported.
    events : Iterable[str]
        The name of the callback events associated to this method.

    See Also
    --------
    Callbackable

    """

    if func is None and len(events) > 0:
        # Decorator called with @callback_method(events=[...]).
        return functools.partial(callback_method, events=events)

    @functools.wraps(func)
    def callbacked_func(self, *args, **kwargs):
        # Store the returned value.
        returned_value = func(self, *args, **kwargs)
        # Invoke callbacks that must be run after the function is executed.
        self._raise_callback_events(func.__name__, *args, **kwargs)
        return returned_value

    # Assign an attribute so that we can distinguish between
    # callbackable and not callbackable methods.
    callbacked_func._callback_events = set(events)

    return callbacked_func


class Callbackable:
    """A base class for registering and handling callbacks.

    The class allows easy implementation of classes that support the
    registration of callback function when some instance methods are
    called. Callback registered to static and class methods are currently
    not supported.

    The class allows callbacks to be registered to single methods or to
    an event, which essentially provides a way to register the same callback
    to multiple methods with a single line.

    Callbacks are called after the function has been executed, and they
    must have the signature

        callback(callbackable, function_name, *args, **kwargs)

    where ``callbackable`` is the ``Callbackable`` instance that raised
    the event, ``function_name`` is the name of the function of ``callbackable``
    that was called, and ``*args`` and ``**kwargs`` are the arguments that
    were passed to the function.

    Examples
    --------

    We define a callbackable list that allows registering callback
    functions to methods that change the list. The first four methods
    summarize the different ways in which the ``callback_method``
    decorator can be used to mark a method for callback registration.

    >>> class CallbackableList(Callbackable, list):
    ...
    ...     @callback_method
    ...     def pop(self, *args, **kwargs):
    ...         super().pop(*args, **kwargs)
    ...
    ...     @callback_method(events=['new_element'])
    ...     def extend(self, *args, **kwargs):
    ...         super().extend(*args, **kwargs)
    ...
    ...     remove = callback_method(list.remove)
    ...     append = callback_method(list.append, events=['new_element'])
    ...
    ...     insert = callback_method(list.insert, events=['new_element'])
    ...     __iadd__ = callback_method(list.__iadd__, events=['new_element'])
    ...     __setitem__ = callback_method(list.__setitem__, events=['new_element'])
    ...     __imul__ = callback_method(list.__imul__)
    ...     __delitem__ = callback_method(list.__delitem__)
    ...

    We can now register a callback to any of the callback methods.

    >>> def callback(callbackable, func_name, *args, **kwargs):
    ...     print(f'callback: {func_name}{args}')
    ...
    >>> l = CallbackableList([1, 2, 3])
    >>> l.register_callback('__delitem__', callback=callback)
    >>> del l[2]
    callback: __delitem__(2,)

    We can also register a callback to a generic event associated to one
    or more methods.

    >>> l.register_callback('new_element', callback=callback)
    >>> l.append(5)
    callback: append(5,)
    >>> l.insert(1, 5)
    callback: insert(1, 5)

    """

    def __init__(self, *args, **kwargs):
        # Map event_name -> list of callbacks.
        self._callbacks = {}
        # Forward arguments to the next class of the MRO to support multiple inheritance.
        super().__init__(*args, **kwargs)

    def register_callback(self, event_name, callback):
        # Check if a callback can be registered for this function. Methods
        # that are not decorated by @callback_method raise an error.
        try:
            attr = getattr(self, event_name)
        except AttributeError:
            # event_name may be an event name. This raises an exception
            # if no method associated to this event is found.
            self._check_event_exist(event_name)
        else:
            # Check if the attribute has been decorated.
            try:
                attr._callback_events
            except AttributeError:
                raise CallbackRegistrationError(
                    f"{self.__class__}.{event_name} is not tagged with the @callback_method decorator"
                )

        # Update the instance callbacks dictionary.
        try:
            self._callbacks[event_name].append(callback)
        except KeyError:
            self._callbacks[event_name] = [callback]

    def _check_event_exist(self, event_name):
        for member in inspect.getmembers(self):
            # If this is not a callback_method, skip.
            try:
                events = member[1]._callback_events
            except AttributeError:
                continue

            # If the event exist, return.
            if event_name in events:
                return

        # The event wasn't found.
        raise CallbackRegistrationError(
            f'No method of {self.__class__} is associated to the callback event "{event_name}".'
        )

    def _raise_callback_events(self, func_name, *args, **kwargs):
        events = getattr(self, func_name)._callback_events

        # Retrieve all the callbacks associated to the function and its events.
        for event_name in [func_name, *events]:
            for callback in self._callbacks.get(event_name, ()):
                callback(self, func_name, *args, **kwargs)


if __name__ == "__main__":
    import doctest

    doctest.run_docstring_examples(Callbackable, globals())



import argparse

from openff.interchange.drivers.openmm import get_openmm_energies
from rdkit.Chem import rdMolAlign

from openff.toolkit import ForceField, Molecule
from openff.toolkit.utils import get_data_file_path


def compute_conformer_energies_from_file(filename):
    # First, load conformers from an SDF file.
    loaded_molecules = Molecule.from_file(
        get_data_file_path("molecules/ruxolitinib_conformers.sdf"),
    )

    # Normalize to list
    try:
        loaded_molecules = [*loaded_molecules]
    except TypeError:
        loaded_molecules = [loaded_molecules]

    # from_file loads each entry in the SDF into its own molecule,
    # so collapse conformers into the same molecule
    molecule = loaded_molecules.pop(0)
    for next_molecule in loaded_molecules:
        if next_molecule == molecule:
            for conformer in next_molecule.conformers:
                molecule.add_conformer(conformer)
        else:
            # We're assuming the SDF just has multiple conformers of the
            # same molecule, so raise an error if that's not the case
            raise ValueError("Multiple chemical species loaded")

    # Make sure the molecule has a name
    if not molecule.name:
        molecule.name = molecule.to_hill_formula()

    print(
        f"Loaded {molecule.n_conformers} conformers"
        + f" of {molecule.to_smiles(explicit_hydrogens=False)!r}"
        + f" ({molecule.name})"
    )

    # Load the openff-2.2.1 force field appropriate for vacuum calculations (without constraints)
    forcefield = ForceField("openff_unconstrained-2.2.1.offxml")
    # Create an Interchange object, which stores the result of parametrizing with this force field
    print(f"Parametrizing {molecule.name} (may take a moment to calculate charges)...")
    interchange = forcefield.create_interchange(molecule.to_topology())
    print("Done.")

    # We'll store energies in two lists
    initial_energies = []
    minimized_energies = []

    # And minimized conformers in a second molecule
    minimized_molecule = Molecule(molecule)
    minimized_molecule.conformers.clear()

    for conformer in molecule.conformers:
        # Use this conformer to update the positions of the Interchange object
        interchange.positions = conformer

        # Get the (total) initial energy from this conformer and store it
        initial_energies.append(get_openmm_energies(interchange).total_energy)

        # Minimize using Interchange.minimize, which wraps OpenMM
        interchange.minimize(engine="openmm")

        # Record minimized energy and positions
        minimized_energies.append(get_openmm_energies(interchange).total_energy)
        minimized_molecule.add_conformer(interchange.positions.to("angstrom"))

    n_confs = molecule.n_conformers
    print(f"{molecule.name}: {n_confs} conformers")

    # Create a copy of the molecule so we can work on it
    working_mol = Molecule(molecule)

    # Print text header
    print("Conformer         Initial PE        Minimized PE        RMSD")
    output = [
        [
            "Conformer",
            "Initial PE (kcal/mol)",
            "Minimized PE (kcal/mol)",
            "RMSD between initial and minimized conformer (Angstrom)",
        ]
    ]

    for i, (init_energy, init_coords, min_energy, min_coords) in enumerate(
        zip(
            initial_energies,
            molecule.conformers,
            minimized_energies,
            minimized_molecule.conformers,
        )
    ):
        # Clear the conformers from the working molecule
        working_mol.conformers.clear()
        # Save the minimized conformer to file
        working_mol.add_conformer(min_coords)
        working_mol.to_file(
            f"{molecule.name}_conf{i + 1}_minimized.sdf",
            file_format="sdf",
        )

        # Calculate the RMSD between the initial and minimized conformer
        working_mol.add_conformer(init_coords)
        rdmol = working_mol.to_rdkit()
        rmslist = []
        rdMolAlign.AlignMolConformers(rdmol, RMSlist=rmslist)
        minimization_rms = rmslist[0]

        # Record the results
        output.append(
            [
                i + 1,
                init_energy.m_as("kilocalories_per_mole"),
                min_energy.m_as("kilocalories_per_mole"),
                minimization_rms,
            ]
        )
        print(f"{{:5d}} / {n_confs:5d} :  {{:8.3f}} kcal/mol {{:8.3f}} kcal/mol {{:8.3f}} Å".format(*output[-1]))

    # Write the results out to CSV
    with open(f"{molecule.name}.csv", "w") as of:
        of.write(", ".join(output.pop(0)) + "\n")
        for line in output:
            of.write("{}, {:.3f}, {:.3f}, {:.3f}".format(*line) + "\n")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description=(
            "Perform energy minimization on a molecule, potentially with many conformers. For each conformer, this "
            "script will provide the initial energy, minimized energy, and RMSD between the initial and minimized "
            "structure (both as STDOUT and a csv file). The minimized conformers will be written out to SDF."
        ),
    )
    parser.add_argument("-f", "--filename", help="Name of an input file containing conformers")
    args = parser.parse_args()

    compute_conformer_energies_from_file(args.filename)



"""
Functions which create a topological molecule directly, without a toolkit.

These are common to several test modules.

"""

import numpy as np

from openff.toolkit import unit
from openff.toolkit._tests.utils import requires_openeye
from openff.toolkit.topology import Molecule, Topology
from openff.toolkit.utils import get_data_file_path


def create_cis_1_2_dichloroethene():
    """
    Creates an openff.toolkit.topology.Molecule representation of cis-1,2-dichloroethene
    without the use of a cheminformatics toolkit.
    """

    cis_dichloroethene = Molecule()
    cis_dichloroethene.add_atom(17, 0, False)
    cis_dichloroethene.add_atom(6, 0, False)
    cis_dichloroethene.add_atom(6, 0, False)
    cis_dichloroethene.add_atom(17, 0, False)
    cis_dichloroethene.add_atom(1, 0, False)
    cis_dichloroethene.add_atom(1, 0, False)
    cis_dichloroethene.add_bond(0, 1, 1, False)
    cis_dichloroethene.add_bond(1, 2, 2, False, "Z")
    cis_dichloroethene.add_bond(2, 3, 1, False)
    cis_dichloroethene.add_bond(1, 4, 1, False)
    cis_dichloroethene.add_bond(2, 5, 1, False)
    return cis_dichloroethene


def create_ethanol():
    """
    Creates an openff.toolkit.topology.Molecule representation of
    ethanol without the use of a cheminformatics toolkit
    """
    # Create an ethanol molecule without using a toolkit
    ethanol = Molecule()
    ethanol.add_atom(6, 0, False)  # C0
    ethanol.add_atom(6, 0, False)  # C1
    ethanol.add_atom(8, 0, False)  # O2
    ethanol.add_atom(1, 0, False)  # H3
    ethanol.add_atom(1, 0, False)  # H4
    ethanol.add_atom(1, 0, False)  # H5
    ethanol.add_atom(1, 0, False)  # H6
    ethanol.add_atom(1, 0, False)  # H7
    ethanol.add_atom(1, 0, False)  # H8
    ethanol.add_bond(0, 1, 1, False, fractional_bond_order=1.33)  # C0 - C1
    ethanol.add_bond(1, 2, 1, False, fractional_bond_order=1.23)  # C1 - O2
    ethanol.add_bond(0, 3, 1, False, fractional_bond_order=1)  # C0 - H3
    ethanol.add_bond(0, 4, 1, False, fractional_bond_order=1)  # C0 - H4
    ethanol.add_bond(0, 5, 1, False, fractional_bond_order=1)  # C0 - H5
    ethanol.add_bond(1, 6, 1, False, fractional_bond_order=1)  # C1 - H6
    ethanol.add_bond(1, 7, 1, False, fractional_bond_order=1)  # C1 - H7
    ethanol.add_bond(2, 8, 1, False, fractional_bond_order=1)  # O2 - H8
    charges = unit.Quantity(
        np.array([-0.4, -0.3, -0.2, -0.1, 0.00001, 0.1, 0.2, 0.3, 0.4]),
        unit.elementary_charge,
    )
    ethanol.partial_charges = charges

    return ethanol


def create_reversed_ethanol():
    """
    Creates an openff.toolkit.topology.Molecule representation of
    ethanol without the use of a cheminformatics toolkit. This function
    reverses the atom indexing of create_ethanol
    """
    # Create an ethanol molecule without using a toolkit
    ethanol = Molecule()
    ethanol.add_atom(1, 0, False)  # H0
    ethanol.add_atom(1, 0, False)  # H1
    ethanol.add_atom(1, 0, False)  # H2
    ethanol.add_atom(1, 0, False)  # H3
    ethanol.add_atom(1, 0, False)  # H4
    ethanol.add_atom(1, 0, False)  # H5
    ethanol.add_atom(8, 0, False)  # O6
    ethanol.add_atom(6, 0, False)  # C7
    ethanol.add_atom(6, 0, False)  # C8
    ethanol.add_bond(8, 7, 1, False, fractional_bond_order=1.33)  # C8 - C7
    ethanol.add_bond(7, 6, 1, False, fractional_bond_order=1.23)  # C7 - O6
    ethanol.add_bond(8, 5, 1, False, fractional_bond_order=1)  # C8 - H5
    ethanol.add_bond(8, 4, 1, False, fractional_bond_order=1)  # C8 - H4
    ethanol.add_bond(8, 3, 1, False, fractional_bond_order=1)  # C8 - H3
    ethanol.add_bond(7, 2, 1, False, fractional_bond_order=1)  # C7 - H2
    ethanol.add_bond(7, 1, 1, False, fractional_bond_order=1)  # C7 - H1
    ethanol.add_bond(6, 0, 1, False, fractional_bond_order=1)  # O6 - H0
    charges = unit.Quantity(
        np.array([0.4, 0.3, 0.2, 0.1, 0.00001, -0.1, -0.2, -0.3, -0.4]),
        unit.elementary_charge,
    )
    ethanol.partial_charges = charges
    return ethanol


def create_benzene_no_aromatic():
    """
    Creates an openff.toolkit.topology.Molecule representation of benzene through the API with aromatic bonds
    not defied, used to test the levels of isomorphic matching.
    """
    benzene = Molecule()
    benzene.add_atom(6, 0, False)  # C0
    benzene.add_atom(6, 0, False)  # C1
    benzene.add_atom(6, 0, False)  # C2
    benzene.add_atom(6, 0, False)  # C3
    benzene.add_atom(6, 0, False)  # C4
    benzene.add_atom(6, 0, False)  # C5
    benzene.add_atom(1, 0, False)  # H6
    benzene.add_atom(1, 0, False)  # H7
    benzene.add_atom(1, 0, False)  # H8
    benzene.add_atom(1, 0, False)  # H9
    benzene.add_atom(1, 0, False)  # H10
    benzene.add_atom(1, 0, False)  # H11
    benzene.add_bond(0, 5, 1, False)  # C0 - C5
    benzene.add_bond(0, 1, 1, False)  # C0 - C1
    benzene.add_bond(1, 2, 1, False)  # C1 - C2
    benzene.add_bond(2, 3, 1, False)  # C2 - C3
    benzene.add_bond(3, 4, 1, False)  # C3 - C4
    benzene.add_bond(4, 5, 1, False)  # C4 - C5
    benzene.add_bond(0, 6, 1, False)  # C0 - H6
    benzene.add_bond(1, 7, 1, False)  # C1 - C7
    benzene.add_bond(2, 8, 1, False)  # C2 - C8
    benzene.add_bond(3, 9, 1, False)  # C3 - C9
    benzene.add_bond(4, 10, 1, False)  # C4 - C10
    benzene.add_bond(5, 11, 1, False)  # C5 - C11
    return benzene


def create_acetaldehyde():
    """
    Creates an openff.toolkit.topology.Molecule representation of acetaldehyde through the API
    """
    acetaldehyde = Molecule()
    acetaldehyde.add_atom(6, 0, False)  # C0
    acetaldehyde.add_atom(6, 0, False)  # C1
    acetaldehyde.add_atom(8, 0, False)  # O2
    acetaldehyde.add_atom(1, 0, False)  # H3
    acetaldehyde.add_atom(1, 0, False)  # H4
    acetaldehyde.add_atom(1, 0, False)  # H5
    acetaldehyde.add_atom(1, 0, False)  # H6
    acetaldehyde.add_bond(0, 1, 1, False)  # C0 - C1
    acetaldehyde.add_bond(1, 2, 2, False)  # C1 = O2
    acetaldehyde.add_bond(0, 3, 1, False)  # C0 - H3
    acetaldehyde.add_bond(0, 4, 1, False)  # C0 - H4
    acetaldehyde.add_bond(0, 5, 1, False)  # C0 - H5
    acetaldehyde.add_bond(1, 6, 1, False)  # C1 - H6
    charges = unit.Quantity(np.zeros(7), unit.elementary_charge)
    acetaldehyde.partial_charges = charges
    return acetaldehyde


def create_water():
    """
    Creates an openff.toolkit.topology.Molecule representation of water through the API
    """
    mol = Molecule()
    mol.add_atom(1, 0, False)  # H1
    mol.add_atom(8, 0, False)  # O
    mol.add_atom(1, 0, False)  # H2
    mol.add_bond(0, 1, 1, False)  # H1 - O
    mol.add_bond(1, 2, 1, False)  # O - H2
    charges = unit.Quantity(np.zeros(3), unit.elementary_charge)
    mol.partial_charges = charges
    return mol


def create_ammonia():
    """
    Creates an openff.toolkit.topology.Molecule representation of ammonia through the API
    """
    mol = Molecule()
    mol.add_atom(1, 0, False)  # H1
    mol.add_atom(7, 0, False)  # N
    mol.add_atom(1, 0, False)  # H2
    mol.add_atom(1, 0, False)  # H3
    mol.add_bond(0, 1, 1, False)  # H1 - N
    mol.add_bond(1, 2, 1, False)  # N - H2
    mol.add_bond(1, 3, 1, False)  # N - H3
    charges = unit.Quantity(np.zeros(4), unit.elementary_charge)
    mol.partial_charges = charges
    return mol


def create_cyclic_n3h3():
    """
    Creates an openff.toolkit.topology.Molecule representation of N3H3 through the API
    """
    mol = Molecule()
    mol.add_atom(7, 0, False)  # N1
    mol.add_atom(7, 0, False)  # N2
    mol.add_atom(7, 0, False)  # N3
    mol.add_atom(1, 0, False)  # H1
    mol.add_atom(1, 0, False)  # H2
    mol.add_atom(1, 0, False)  # H3
    mol.add_bond(0, 1, 1, False)  # N1 - N2
    mol.add_bond(1, 2, 1, False)  # N2 - N3
    mol.add_bond(2, 0, 1, False)  # N1 - N3
    mol.add_bond(0, 3, 1, False)  # N1 - H1
    mol.add_bond(1, 4, 1, False)  # N2 - H2
    mol.add_bond(2, 5, 1, False)  # N3 - H3
    return mol


def create_acetate():
    """
    Creates an openff.toolkit.topology.Molecule representation of
    acetate without the use of a cheminformatics toolkit
    """
    # Create an acetate molecule without using a toolkit
    acetate = Molecule()
    acetate.add_atom(6, 0, False)  # C0
    acetate.add_atom(6, 0, False)  # C1
    acetate.add_atom(8, 0, False)  # O2
    acetate.add_atom(8, -1, False)  # O3
    acetate.add_atom(1, 0, False)  # H4
    acetate.add_atom(1, 0, False)  # H5
    acetate.add_atom(1, 0, False)  # H6
    acetate.add_bond(0, 1, 1, False)  # C0 - C1
    acetate.add_bond(1, 2, 2, False)  # C1 = O2
    acetate.add_bond(1, 3, 1, False)  # C1 - O3[-1]
    acetate.add_bond(0, 4, 1, False)  # C0 - H4
    acetate.add_bond(0, 5, 1, False)  # C0 - H5
    acetate.add_bond(0, 6, 1, False)  # C0 - H6
    return acetate


def create_cyclohexane():
    """
    Creates an openff.toolkit.topology.Molecule representation of
    cyclohexane without the use of a cheminformatics toolkit
    """
    cyclohexane = Molecule()
    cyclohexane.add_atom(6, 0, False)  # C0
    cyclohexane.add_atom(6, 0, False)  # C1
    cyclohexane.add_atom(6, 0, False)  # C2
    cyclohexane.add_atom(6, 0, False)  # C3
    cyclohexane.add_atom(6, 0, False)  # C4
    cyclohexane.add_atom(6, 0, False)  # C5
    cyclohexane.add_atom(1, 0, False)  # H6
    cyclohexane.add_atom(1, 0, False)  # H7
    cyclohexane.add_atom(1, 0, False)  # H8
    cyclohexane.add_atom(1, 0, False)  # H9
    cyclohexane.add_atom(1, 0, False)  # H10
    cyclohexane.add_atom(1, 0, False)  # H11
    cyclohexane.add_atom(1, 0, False)  # H12
    cyclohexane.add_atom(1, 0, False)  # H13
    cyclohexane.add_atom(1, 0, False)  # H14
    cyclohexane.add_atom(1, 0, False)  # H15
    cyclohexane.add_atom(1, 0, False)  # H16
    cyclohexane.add_atom(1, 0, False)  # H17
    cyclohexane.add_bond(0, 1, 1, False)  # C0 - C1
    cyclohexane.add_bond(1, 2, 1, False)  # C1 - C2
    cyclohexane.add_bond(2, 3, 1, False)  # C2 - C3
    cyclohexane.add_bond(3, 4, 1, False)  # C3 - C4
    cyclohexane.add_bond(4, 5, 1, False)  # C4 - C5
    cyclohexane.add_bond(5, 0, 1, False)  # C5 - C0
    cyclohexane.add_bond(0, 6, 1, False)  # C0 - H6
    cyclohexane.add_bond(0, 7, 1, False)  # C0 - H7
    cyclohexane.add_bond(1, 8, 1, False)  # C1 - H8
    cyclohexane.add_bond(1, 9, 1, False)  # C1 - H9
    cyclohexane.add_bond(2, 10, 1, False)  # C2 - H10
    cyclohexane.add_bond(2, 11, 1, False)  # C2 - H11
    cyclohexane.add_bond(3, 12, 1, False)  # C3 - H12
    cyclohexane.add_bond(3, 13, 1, False)  # C3 - H13
    cyclohexane.add_bond(4, 14, 1, False)  # C4 - H14
    cyclohexane.add_bond(4, 15, 1, False)  # C4 - H15
    cyclohexane.add_bond(5, 16, 1, False)  # C5 - H16
    cyclohexane.add_bond(5, 17, 1, False)  # C5 - H17
    return cyclohexane


def create_dioxygen():
    """
    Creates an openff.toolkit.topology.Molecule representation of
    dioxygen without the use of a cheminformatics toolkit
    """
    dioxygen = Molecule()
    dioxygen.add_atom(8, 0, False)  # O0
    dioxygen.add_atom(8, 0, False)  # O1
    dioxygen.add_bond(0, 1, 2, False)  # O0 # O1
    charges = unit.Quantity(np.array([0.0, 0.0]), unit.elementary_charge)
    dioxygen.partial_charges = charges

    return dioxygen


def create_dinitrogen():
    """
    Creates an openff.toolkit.topology.Molecule representation of
    dinitrogen without the use of a cheminformatics toolkit
    """
    dinitrogen = Molecule()
    dinitrogen.add_atom(7, 0, False)  # N0
    dinitrogen.add_atom(7, 0, False)  # N1
    dinitrogen.add_bond(0, 1, 3, False)  # N0 - N1
    charges = unit.Quantity(np.array([0.0, 0.0]), unit.elementary_charge)
    dinitrogen.partial_charges = charges
    return dinitrogen


def dipeptide():
    dipeptide = Molecule.from_file(get_data_file_path("proteins/CTerminal_ALA.sdf"))
    return dipeptide


def dipeptide_residues_perceived():
    dipeptide_residues_perceived = Molecule(dipeptide())
    dipeptide_residues_perceived.perceive_residues()
    return dipeptide_residues_perceived


def dipeptide_hierarchy_added():
    dipeptide_hierarchy_perceived = Molecule(dipeptide_residues_perceived())
    dipeptide_hierarchy_perceived.add_default_hierarchy_schemes()
    return dipeptide_hierarchy_perceived


def cyx():
    cyx = Molecule.from_file(get_data_file_path("proteins/MainChain_CYX.sdf"))
    return cyx


def cyx_residues_perceived():
    cyx_residues_perceived = Molecule(cyx())
    cyx_residues_perceived.perceive_residues()
    return cyx_residues_perceived


def cyx_hierarchy_added():
    cyx_hierarchy_perceived = Molecule(cyx_residues_perceived())
    cyx_hierarchy_perceived.add_default_hierarchy_schemes()
    return cyx_hierarchy_perceived


def empty_molecule():
    return Molecule()


def ethane_from_smiles():
    return Molecule.from_smiles("CC")


def ethene_from_smiles():
    return Molecule.from_smiles("C=C")


def propane_from_smiles():
    return Molecule.from_smiles("CCC")


def toluene_from_sdf():
    filename = get_data_file_path("molecules/toluene.sdf")
    return Molecule.from_file(filename)


@requires_openeye
def toluene_from_charged_mol2():
    filename = get_data_file_path("molecules/toluene_charged.mol2")
    # TODO: This will require openeye to load
    return Molecule.from_file(filename)


def charged_methylamine_from_smiles():
    return Molecule.from_smiles("[H]C([H])([H])[N+]([H])([H])[H]")


def ethane_from_smiles_w_vsites():
    molecule = Molecule.from_smiles("CC")
    carbons = [atom for atom in molecule.atoms if atom.atomic_number == 6]
    c0_hydrogens = [atom for atom in carbons[0].bonded_atoms if atom.atomic_number == 1]
    molecule.add_bond_charge_virtual_site(
        (carbons[0], carbons[1]),
        0.1 * unit.angstrom,
        charge_increments=[0.1, 0.05] * unit.elementary_charge,
    )
    molecule.add_monovalent_lone_pair_virtual_site(
        (c0_hydrogens[0], carbons[0], carbons[1]),
        0.2 * unit.angstrom,
        20 * unit.degree,
        25 * unit.degree,
        charge_increments=[0.01, 0.02, 0.03] * unit.elementary_charge,
    )
    return molecule


def propane_from_smiles_w_vsites():
    # Make a propane with virtual sites
    molecule = Molecule.from_smiles("CCC")
    carbons = [atom for atom in molecule.atoms if atom.atomic_number == 6]
    c0_hydrogens = [atom for atom in carbons[0].bonded_atoms if atom.atomic_number == 1]
    # This will add *two* particles (symmetric=True), *one* virtual site
    molecule.add_monovalent_lone_pair_virtual_site(
        (c0_hydrogens[0], carbons[0], carbons[1]),
        0.2 * unit.angstrom,
        20 * unit.degree,
        25 * unit.degree,
        charge_increments=[0.01, 0.02, 0.03] * unit.elementary_charge,
        symmetric=True,
    )
    # This will add *one* particle (symmetric=False), *one* virtual site
    molecule.add_bond_charge_virtual_site(
        (carbons[0], carbons[1]),
        0.1 * unit.angstrom,
        charge_increments=[0.1, 0.05] * unit.elementary_charge,
        symmetric=False,
    )
    return molecule


def tip5_water():
    # Make a TIP5 water
    molecule = Molecule.from_smiles("[H][O][H]")
    O1 = next(atom for atom in molecule.atoms if atom.atomic_number == 8)
    H1, H2 = (atom for atom in O1.bonded_atoms if atom.atomic_number == 1)
    molecule.add_divalent_lone_pair_virtual_site(
        (H1, O1, H2),
        0.7 * unit.angstrom,
        54.71384225 * unit.degree,
        charge_increments=[0.1205, 0.00, 0.1205] * unit.elementary_charge,
        symmetric=True,
    )
    return molecule


def topology_with_metadata():
    n2 = Molecule.from_smiles("N#N")
    ammonia = create_ammonia()
    chloride = Molecule.from_smiles("[Cl-]")
    water = create_water()
    mols = [
        n2,
        ammonia,
        chloride,
        chloride,
        ammonia,
        chloride,
        chloride,
        n2,
        chloride,
        chloride,
        chloride,
        water,
    ]

    top = Topology.from_molecules(mols)

    atom_hier_info = [
        ("AAA", 1, " ", "A"),  # mol[0]
        ("AAA", 1, " ", "A"),  # mol[0]
        ("AAA", 1, " ", "A"),  # mol[1]
        ("BBB", 1, " ", "A"),  # mol[1]
        ("BBB", 2, " ", "A"),  # mol[1]
        ("BBB", 2, " ", "B"),  # mol[1]
        ("CCC", 2, " ", "B"),  # mol[2]
        ("CCC", 3, " ", "B"),  # mol[3]
        ("CCC", 3, " ", "C"),  # mol[4]
        ("DDD", 4, " ", "C"),  # mol[4]
        ("EEE", 4, " ", "D"),  # mol[4]
        ("EEE", 5, " ", "E"),  # mol[4]
        ("FFF", 6, " ", "E"),  # mol[5]
        ("GGG", 6, " ", "F"),  # mol[6]
        ("GGG", 7, " ", "G"),  # mol[7]
        ("HHH", 8, " ", "H"),  # mol[7]
        ("YZ", 9, " ", "I"),  # mol[8]
        ("AAA", 1, " ", "A"),  # mol[9]
        ("AAA", 1, "X", "A"),  # mol[10]
        (None, None, " ", None),  # mol[11]
        (None, None, " ", None),  # mol[12]
        ("AAA", None, " ", None),  # mol[13]
        (None, None, " ", "A"),  # mol[14]
    ]

    for atom, metadata_tuple in zip(top.atoms, atom_hier_info):
        residue_name = metadata_tuple[0]
        residue_number = metadata_tuple[1]
        insertion_code = metadata_tuple[2]
        chain_id = metadata_tuple[3]
        if residue_name is not None:
            atom.metadata["residue_name"] = residue_name
        if residue_number is not None:
            atom.metadata["residue_number"] = residue_number
        if insertion_code is not None:
            atom.metadata["insertion_code"] = insertion_code
        if chain_id is not None:
            atom.metadata["chain_id"] = chain_id

    return top



"""
Configuration file for pytest.
"""

import logging

import pytest

logger = logging.getLogger(__name__)

# Ensure QCPortal is imported before any OpenEye modules, see
# https://github.com/conda-forge/qcfractal-feedstock/issues/43
try:
    import qcportal  # noqa
except ImportError:
    pass


def untar_full_alkethoh_and_freesolv_set():
    """When running slow tests, we unpack the full AlkEthOH and FreeSolv test
    sets in advance to speed things up.

    See
        test_forcefield.py::test_alkethoh_parameters_assignment
        test_forcefield.py::test_freesolv_parameters_assignment
    """
    import os
    import tarfile

    from openff.toolkit.utils import get_data_file_path

    molecule_dir_path = get_data_file_path("molecules")
    for tarfile_name in ["AlkEthOH_tripos.tar.gz", "FreeSolv.tar.gz"]:
        tarfile_path = os.path.join(molecule_dir_path, tarfile_name)
        with tarfile.open(tarfile_path, "r:gz") as tar:
            tar.extractall(path=molecule_dir_path)


def pytest_addoption(parser):
    """Add the pytest command line option --runslow and --failwip.

    If --runslow is not given, tests marked with pytest.mark.slow are
    skipped.

    If --failwip is not given, tests marked with pytest.mark.wip are
    xfailed.

    Parameters
    ----------
    parser : argparsing.parser
        The parser used by pytest to process arguments

    Returns
    -------
    None

    """

    # Loaded pytest plugins define their own arguments, and in certain cases
    # our options use the same name. Although this can define two different
    # behaviors for the same argument, the two options below are fairly
    # unambiguous in their interpretation, so we skip any errors and allow
    # the testing to proceed

    try:
        parser.addoption("--runslow", action="store_true", default=False, help="run slow tests")
    except ValueError:
        logger.warning("Option --runslow already added elsewhere (from a plugin possibly?). Skipping...")

    try:
        parser.addoption(
            "--failwip",
            action="store_true",
            default=False,
            help="fail work in progress tests",
        )
    except ValueError:
        logger.warning("Option --failwip already added elsewhere (from a plugin possibly?). Skipping...")


def pytest_collection_modifyitems(config, items):
    if config.getoption("runslow"):
        # If --runslow is given, we don't have to mark items for skipping,
        # but we need to extract the whole AlkEthOH and FreeSolv sets (see
        # test_forcefield::test_alkethoh/freesolv_parameters_assignment).
        untar_full_alkethoh_and_freesolv_set()
    else:
        # Mark for skipping all items marked as slow.
        skip_slow = pytest.mark.skip(reason="specify --runslow pytest option to run this test.")
        for item in items:
            if "slow" in item.keywords:
                item.add_marker(skip_slow)

    # Mark work-in-progress tests for xfail.
    if not config.getoption("failwip"):
        xfail_wip_reason = "This is a work in progress test. Specify --failwip pytest option to make this test fail."
        for item in items:
            if "wip" in item.keywords:
                # Augment original reason.
                reason = xfail_wip_reason + item.get_closest_marker("wip").kwargs.get("reason", "")
                item.add_marker(pytest.mark.xfail(reason=reason))


@pytest.fixture
def opc():
    from openff.toolkit import ForceField

    return ForceField("opc.offxml")
